@@ -1,7 +1,10 @@
 import ast
+import collections.abc as collections_abc
+import inspect
 import keyword
 import re
-from typing import TYPE_CHECKING, Any, Dict, Iterable, List
+from enum import Enum
+from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Optional
 
 from prompt_toolkit.completion import (
     CompleteEvent,
@@ -9,16 +12,28 @@ from prompt_toolkit.completion import (
     Completion,
     PathCompleter,
 )
+from prompt_toolkit.contrib.completers.system import SystemCompleter
 from prompt_toolkit.contrib.regular_languages.compiler import compile as compile_grammar
 from prompt_toolkit.contrib.regular_languages.completion import GrammarCompleter
 from prompt_toolkit.document import Document
+from prompt_toolkit.formatted_text import fragment_list_to_text, to_formatted_text
 
 from ptpython.utils import get_jedi_script_from_document
 
 if TYPE_CHECKING:
     from prompt_toolkit.contrib.regular_languages.compiler import _CompiledGrammar
 
-__all__ = ["PythonCompleter"]
+__all__ = ["PythonCompleter", "CompletePrivateAttributes", "HidePrivateCompleter"]
+
+
+class CompletePrivateAttributes(Enum):
+    """
+    Should we display private attributes in the completion pop-up?
+    """
+
+    NEVER = "NEVER"
+    IF_NO_PUBLIC = "IF_NO_PUBLIC"
+    ALWAYS = "ALWAYS"
 
 
 class PythonCompleter(Completer):
@@ -26,17 +41,20 @@ class PythonCompleter(Completer):
     Completer for Python code.
     """
 
-    def __init__(self, get_globals, get_locals, get_enable_dictionary_completion):
+    def __init__(
+        self, get_globals, get_locals, get_enable_dictionary_completion
+    ) -> None:
         super().__init__()
 
         self.get_globals = get_globals
         self.get_locals = get_locals
         self.get_enable_dictionary_completion = get_enable_dictionary_completion
 
-        self.dictionary_completer = DictionaryCompleter(get_globals, get_locals)
+        self._system_completer = SystemCompleter()
+        self._dictionary_completer = DictionaryCompleter(get_globals, get_locals)
 
-        self._path_completer_cache = None
-        self._path_completer_grammar_cache = None
+        self._path_completer_cache: Optional[GrammarCompleter] = None
+        self._path_completer_grammar_cache: Optional["_CompiledGrammar"] = None
 
     @property
     def _path_completer(self) -> GrammarCompleter:
@@ -123,10 +141,20 @@ class PythonCompleter(Completer):
         """
         Get Python completions.
         """
+        # If the input starts with an exclamation mark. Use the system completer.
+        if document.text.lstrip().startswith("!"):
+            yield from self._system_completer.get_completions(
+                Document(
+                    text=document.text[1:], cursor_position=document.cursor_position - 1
+                ),
+                complete_event,
+            )
+            return
+
         # Do dictionary key completions.
         if self.get_enable_dictionary_completion():
             has_dict_completions = False
-            for c in self.dictionary_completer.get_completions(
+            for c in self._dictionary_completer.get_completions(
                 document, complete_event
             ):
                 if c.text not in "[.":
@@ -141,8 +169,7 @@ class PythonCompleter(Completer):
         if complete_event.completion_requested or self._complete_path_while_typing(
             document
         ):
-            for c in self._path_completer.get_completions(document, complete_event):
-                yield c
+            yield from self._path_completer.get_completions(document, complete_event)
 
         # If we are inside a string, Don't do Jedi completion.
         if self._path_completer_grammar.match(document.text_before_cursor):
@@ -158,7 +185,10 @@ class PythonCompleter(Completer):
 
             if script:
                 try:
-                    completions = script.completions()
+                    jedi_completions = script.complete(
+                        column=document.cursor_position_col,
+                        line=document.cursor_position_row + 1,
+                    )
                 except TypeError:
                     # Issue #9: bad syntax causes completions() to fail in jedi.
                     # https://github.com/jonathanslenders/python-prompt-toolkit/issues/9
@@ -196,12 +226,18 @@ class PythonCompleter(Completer):
                     # Supress all other Jedi exceptions.
                     pass
                 else:
-                    for c in completions:
+                    for jc in jedi_completions:
+                        if jc.type == "function":
+                            suffix = "()"
+                        else:
+                            suffix = ""
+
                         yield Completion(
-                            c.name_with_symbols,
-                            len(c.complete) - len(c.name_with_symbols),
-                            display=c.name_with_symbols,
-                            style=_get_style_for_name(c.name_with_symbols),
+                            jc.name_with_symbols,
+                            len(jc.complete) - len(jc.name_with_symbols),
+                            display=jc.name_with_symbols + suffix,
+                            display_meta=jc.type,
+                            style=_get_style_for_name(jc.name_with_symbols),
                         )
 
 
@@ -354,9 +390,14 @@ class DictionaryCompleter(Completer):
             object_var = match.groups()[0]
             result = self._lookup(object_var, temp_locals)
 
-            if isinstance(result, (list, tuple, dict)):
+            if isinstance(
+                result,
+                (list, tuple, dict, collections_abc.Mapping, collections_abc.Sequence),
+            ):
                 yield Completion("[", 0)
-            elif result:
+            elif result is not None:
+                # Note: Don't call `if result` here. That can fail for types
+                #       that have custom truthness checks.
                 yield Completion(".", 0)
 
     def _get_item_lookup_completions(
@@ -368,6 +409,16 @@ class DictionaryCompleter(Completer):
         """
         Complete dictionary keys.
         """
+
+        def abbr_meta(text: str) -> str:
+            " Abbreviate meta text, make sure it fits on one line. "
+            # Take first line, if multiple lines.
+            if len(text) > 20:
+                text = text[:20] + "..."
+            if "\n" in text:
+                text = text.split("\n", 1)[0] + "..."
+            return text
+
         match = self.item_lookup_pattern.search(document.text_before_cursor)
         if match is not None:
             object_var, key = match.groups()
@@ -376,7 +427,7 @@ class DictionaryCompleter(Completer):
             result = self._lookup(object_var, temp_locals)
 
             # If this object is a dictionary, complete the keys.
-            if isinstance(result, dict):
+            if isinstance(result, (dict, collections_abc.Mapping)):
                 # Try to evaluate the key.
                 key_obj = key
                 for k in [key, key + '"', key + "'"]:
@@ -395,13 +446,13 @@ class DictionaryCompleter(Completer):
                                 k_repr + "]",
                                 -len(key),
                                 display=f"[{k_repr}]",
-                                display_meta=self._do_repr(result[k]),
+                                display_meta=abbr_meta(self._do_repr(result[k])),
                             )
                         except ReprFailedError:
                             pass
 
             # Complete list/tuple index keys.
-            elif isinstance(result, (list, tuple)):
+            elif isinstance(result, (list, tuple, collections_abc.Sequence)):
                 if not key or key.isdigit():
                     for k in range(min(len(result), 1000)):
                         if str(k).startswith(key):
@@ -411,7 +462,7 @@ class DictionaryCompleter(Completer):
                                     k_repr + "]",
                                     -len(key),
                                     display=f"[{k_repr}]",
-                                    display_meta=self._do_repr(result[k]),
+                                    display_meta=abbr_meta(self._do_repr(result[k])),
                                 )
                             except ReprFailedError:
                                 pass
@@ -434,9 +485,24 @@ class DictionaryCompleter(Completer):
 
             names = self._sort_attribute_names(dir(result))
 
+            def get_suffix(name: str) -> str:
+                try:
+                    obj = getattr(result, name, None)
+                    if inspect.isfunction(obj):
+                        return "()"
+
+                    if isinstance(obj, dict):
+                        return "{}"
+                    if isinstance(obj, (list, tuple)):
+                        return "[]"
+                except:
+                    pass
+                return ""
+
             for name in names:
                 if name.startswith(attr_name):
-                    yield Completion(name, -len(attr_name))
+                    suffix = get_suffix(name)
+                    yield Completion(name, -len(attr_name), display=name + suffix)
 
     def _sort_attribute_names(self, names: List[str]) -> List[str]:
         """
@@ -454,6 +520,49 @@ class DictionaryCompleter(Completer):
         return sorted(names, key=sort_key)
 
 
+class HidePrivateCompleter(Completer):
+    """
+    Wrapper around completer that hides private fields, deponding on whether or
+    not public fields are shown.
+
+    (The reason this is implemented as a `Completer` wrapper is because this
+    way it works also with `FuzzyCompleter`.)
+    """
+
+    def __init__(
+        self,
+        completer: Completer,
+        complete_private_attributes: Callable[[], CompletePrivateAttributes],
+    ) -> None:
+        self.completer = completer
+        self.complete_private_attributes = complete_private_attributes
+
+    def get_completions(
+        self, document: Document, complete_event: CompleteEvent
+    ) -> Iterable[Completion]:
+
+        completions = list(self.completer.get_completions(document, complete_event))
+        complete_private_attributes = self.complete_private_attributes()
+        hide_private = False
+
+        def is_private(completion: Completion) -> bool:
+            text = fragment_list_to_text(to_formatted_text(completion.display))
+            return text.startswith("_")
+
+        if complete_private_attributes == CompletePrivateAttributes.NEVER:
+            hide_private = True
+
+        elif complete_private_attributes == CompletePrivateAttributes.IF_NO_PUBLIC:
+            hide_private = any(not is_private(completion) for completion in completions)
+
+        if hide_private:
+            completions = [
+                completion for completion in completions if not is_private(completion)
+            ]
+
+        return completions
+
+
 class ReprFailedError(Exception):
     " Raised when the repr() call in `DictionaryCompleter` fails. "
 
