@@ -11,15 +11,17 @@ import asyncio
 import builtins
 import os
 import sys
+import threading
 import traceback
+import types
 import warnings
+from dis import COMPILER_FLAG_NAMES
 from enum import Enum
 from typing import Any, Callable, ContextManager, Dict, Optional
 
-import black
-from prompt_toolkit.document import Document
 from prompt_toolkit.formatted_text import (
     HTML,
+    AnyFormattedText,
     FormattedText,
     PygmentsTokens,
     StyleAndTextTuples,
@@ -29,7 +31,6 @@ from prompt_toolkit.formatted_text import (
 )
 from prompt_toolkit.formatted_text.utils import fragment_list_to_text, split_lines
 from prompt_toolkit.key_binding import KeyBindings, KeyPressEvent
-from prompt_toolkit.key_binding.vi_state import InputMode
 from prompt_toolkit.patch_stdout import patch_stdout as patch_stdout_context
 from prompt_toolkit.shortcuts import (
     PromptSession,
@@ -44,15 +45,39 @@ from pygments.token import Token
 
 from .python_input import PythonInput
 
+try:
+    from ast import PyCF_ALLOW_TOP_LEVEL_AWAIT  # type: ignore
+except ImportError:
+    PyCF_ALLOW_TOP_LEVEL_AWAIT = 0
+
 __all__ = ["PythonRepl", "enable_deprecation_warnings", "run_config", "embed"]
 
 
+def _get_coroutine_flag() -> Optional[int]:
+    for k, v in COMPILER_FLAG_NAMES.items():
+        if v == "COROUTINE":
+            return k
+
+    # Flag not found.
+    return None
+
+
+COROUTINE_FLAG: Optional[int] = _get_coroutine_flag()
+
+
+def _has_coroutine_flag(code: types.CodeType) -> bool:
+    if COROUTINE_FLAG is None:
+        # Not supported on this Python version.
+        return False
+
+    return bool(code.co_flags & COROUTINE_FLAG)
+
+
 class PythonRepl(PythonInput):
     def __init__(self, *a, **kw) -> None:
         self._startup_paths = kw.pop("startup_paths", None)
         super().__init__(*a, **kw)
         self._load_start_paths()
-        self.pt_loop = asyncio.new_event_loop()
 
     def _load_start_paths(self) -> None:
         " Start the Read-Eval-Print Loop. "
@@ -67,77 +92,97 @@ class PythonRepl(PythonInput):
                     output.write("WARNING | File not found: {}\n\n".format(path))
 
     def run(self) -> None:
-        # In order to make sure that asyncio code written in the
-        # interactive shell doesn't interfere with the prompt, we run the
-        # prompt in a different event loop.
-        # If we don't do this, people could spawn coroutine with a
-        # while/true inside which will freeze the prompt.
-
-        try:
-            old_loop: Optional[asyncio.AbstractEventLoop] = asyncio.get_event_loop()
-        except RuntimeError:
-            # This happens when the user used `asyncio.run()`.
-            old_loop = None
-
-        asyncio.set_event_loop(self.pt_loop)
-        try:
-            return self.pt_loop.run_until_complete(self.run_async())
-        finally:
-            # Restore the original event loop.
-            asyncio.set_event_loop(old_loop)
-
-    async def run_async(self) -> None:
+        """
+        Run the REPL loop.
+        """
         if self.terminal_title:
             set_title(self.terminal_title)
 
-        while True:
-            # Capture the current input_mode in order to restore it after reset,
-            # for ViState.reset() sets it to InputMode.INSERT unconditionally and
-            # doesn't accept any arguments.
-            def pre_run(
-                last_input_mode: InputMode = self.app.vi_state.input_mode,
-            ) -> None:
-                if self.vi_keep_last_used_mode:
-                    self.app.vi_state.input_mode = last_input_mode
+        self._add_to_namespace()
 
-                if not self.vi_keep_last_used_mode and self.vi_start_in_navigation_mode:
-                    self.app.vi_state.input_mode = InputMode.NAVIGATION
+        try:
+            while True:
+                # Read.
+                try:
+                    text = self.read()
+                except EOFError:
+                    return
 
-            # Run the UI.
-            try:
-                text = await self.app.run_async(pre_run=pre_run)
-            except EOFError:
-                return
-            except KeyboardInterrupt:
-                # Abort - try again.
-                self.default_buffer.document = Document()
-            else:
-                await self._process_text(text)
+                # Eval.
+                try:
+                    result = self.eval(text)
+                except KeyboardInterrupt as e:  # KeyboardInterrupt doesn't inherit from Exception.
+                    self._handle_keyboard_interrupt(e)
+                except SystemExit:
+                    return
+                except BaseException as e:
+                    self._handle_exception(e)
+                else:
+                    # Print.
+                    if result is not None:
+                        self.show_result(result)
+
+                    # Loop.
+                    self.current_statement_index += 1
+                    self.signatures = []
+
+        finally:
+            if self.terminal_title:
+                clear_title()
+            self._remove_from_namespace()
+
+    async def run_async(self) -> None:
+        """
+        Run the REPL loop, but run the blocking parts in an executor, so that
+        we don't block the event loop. Both the input and output (which can
+        display a pager) will run in a separate thread with their own event
+        loop, this way ptpython's own event loop won't interfere with the
+        asyncio event loop from where this is called.
+
+        The "eval" however happens in the current thread, which is important.
+        (Both for control-C to work, as well as for the code to see the right
+        thread in which it was embedded).
+        """
+        loop = asyncio.get_event_loop()
 
         if self.terminal_title:
-            clear_title()
+            set_title(self.terminal_title)
 
-    async def _process_text(self, line: str) -> None:
+        self._add_to_namespace()
 
-        if line and not line.isspace():
-            if self.insert_blank_line_after_input:
-                self.app.output.write("\n")
+        try:
+            while True:
+                # Read.
+                try:
+                    text = await loop.run_in_executor(None, self.read)
+                except EOFError:
+                    return
 
-            try:
-                # Eval and print.
-                await self._execute(line)
-            except KeyboardInterrupt as e:  # KeyboardInterrupt doesn't inherit from Exception.
-                self._handle_keyboard_interrupt(e)
-            except Exception as e:
-                self._handle_exception(e)
+                # Eval.
+                try:
+                    result = await self.eval_async(text)
+                except KeyboardInterrupt as e:  # KeyboardInterrupt doesn't inherit from Exception.
+                    self._handle_keyboard_interrupt(e)
+                except SystemExit:
+                    return
+                except BaseException as e:
+                    self._handle_exception(e)
+                else:
+                    # Print.
+                    if result is not None:
+                        await loop.run_in_executor(
+                            None, lambda: self.show_result(result)
+                        )
 
-            if self.insert_blank_line_after_output:
-                self.app.output.write("\n")
+                    # Loop.
+                    self.current_statement_index += 1
+                    self.signatures = []
+        finally:
+            if self.terminal_title:
+                clear_title()
+            self._remove_from_namespace()
 
-            self.current_statement_index += 1
-            self.signatures = []
-
-    async def _execute(self, line: str) -> None:
+    def eval(self, line: str) -> object:
         """
         Evaluate the line and print the result.
         """
@@ -146,45 +191,79 @@ class PythonRepl(PythonInput):
         if "" not in sys.path:
             sys.path.insert(0, "")
 
-        def compile_with_flags(code: str, mode: str):
-            " Compile code with the right compiler flags. "
-            return compile(
-                code,
-                "<stdin>",
-                mode,
-                flags=self.get_compiler_flags(),
-                dont_inherit=True,
-            )
-
-        # If the input is single line, remove leading whitespace.
-        # (This doesn't have to be a syntax error.)
-        if len(line.splitlines()) == 1:
-            line = line.strip()
-
-        if line.lstrip().startswith("\x1a"):
-            # When the input starts with Ctrl-Z, quit the REPL.
-            self.app.exit()
-
-        elif line.lstrip().startswith("!"):
+        if line.lstrip().startswith("!"):
             # Run as shell command
             os.system(line[1:])
         else:
             # Try eval first
             try:
-                code = compile_with_flags(line, "eval")
+                code = self._compile_with_flags(line, "eval")
+            except SyntaxError:
+                # If not a valid `eval` expression, run using `exec` instead.
+                code = self._compile_with_flags(line, "exec")
+                exec(code, self.get_globals(), self.get_locals())
+            else:
+                # No syntax errors for eval. Do eval.
                 result = eval(code, self.get_globals(), self.get_locals())
 
-                locals: Dict[str, Any] = self.get_locals()
-                locals["_"] = locals["_%i" % self.current_statement_index] = result
+                if _has_coroutine_flag(code):
+                    result = asyncio.get_event_loop().run_until_complete(result)
 
-                if result is not None:
-                    await self.show_result(result)
-            # If not a valid `eval` expression, run using `exec` instead.
+                self._store_eval_result(result)
+                return result
+
+        return None
+
+    async def eval_async(self, line: str) -> object:
+        """
+        Evaluate the line and print the result.
+        """
+        # WORKAROUND: Due to a bug in Jedi, the current directory is removed
+        # from sys.path. See: https://github.com/davidhalter/jedi/issues/1148
+        if "" not in sys.path:
+            sys.path.insert(0, "")
+
+        if line.lstrip().startswith("!"):
+            # Run as shell command
+            os.system(line[1:])
+        else:
+            # Try eval first
+            try:
+                code = self._compile_with_flags(line, "eval")
             except SyntaxError:
-                code = compile_with_flags(line, "exec")
+                # If not a valid `eval` expression, run using `exec` instead.
+                code = self._compile_with_flags(line, "exec")
                 exec(code, self.get_globals(), self.get_locals())
+            else:
+                # No syntax errors for eval. Do eval.
+                result = eval(code, self.get_globals(), self.get_locals())
 
-    async def show_result(self, result: object) -> None:
+                if _has_coroutine_flag(code):
+                    result = await result
+
+                self._store_eval_result(result)
+                return result
+
+        return None
+
+    def _store_eval_result(self, result: object) -> None:
+        locals: Dict[str, Any] = self.get_locals()
+        locals["_"] = locals["_%i" % self.current_statement_index] = result
+
+    def get_compiler_flags(self) -> int:
+        return super().get_compiler_flags() | PyCF_ALLOW_TOP_LEVEL_AWAIT
+
+    def _compile_with_flags(self, code: str, mode: str):
+        " Compile code with the right compiler flags. "
+        return compile(
+            code,
+            "<stdin>",
+            mode,
+            flags=self.get_compiler_flags(),
+            dont_inherit=True,
+        )
+
+    def show_result(self, result: object) -> None:
         """
         Show __repr__ for an `eval` result.
         """
@@ -199,6 +278,10 @@ class PythonRepl(PythonInput):
         else:
             # Syntactically correct. Format with black and syntax highlight.
             if self.enable_output_formatting:
+                # Inline import. Slightly speed up start-up time if black is
+                # not used.
+                import black
+
                 result_repr = black.format_str(
                     result_repr,
                     mode=black.FileMode(line_length=self.app.output.get_size().columns),
@@ -210,15 +293,17 @@ class PythonRepl(PythonInput):
 
         # If __pt_repr__ is present, take this. This can return prompt_toolkit
         # formatted text.
-        if hasattr(result, "__pt_repr__"):
-            try:
+        try:
+            if hasattr(result, "__pt_repr__"):
                 formatted_result_repr = to_formatted_text(
                     getattr(result, "__pt_repr__")()
                 )
                 if isinstance(formatted_result_repr, list):
                     formatted_result_repr = FormattedText(formatted_result_repr)
-            except:
-                pass
+        except:
+            # For bad code, `__getattr__` can raise something that's not an
+            # `AttributeError`. This happens already when calling `hasattr()`.
+            pass
 
         # Align every line to the prompt.
         line_sep = "\n" + " " * fragment_list_width(out_prompt)
@@ -242,14 +327,15 @@ class PythonRepl(PythonInput):
             )
 
         if self.enable_pager:
-            await self._print_paginated_formatted_text(
-                to_formatted_text(formatted_output)
-            )
+            self.print_paginated_formatted_text(to_formatted_text(formatted_output))
         else:
             self.print_formatted_text(to_formatted_text(formatted_output))
 
         self.app.output.flush()
 
+        if self.insert_blank_line_after_output:
+            self.app.output.write("\n")
+
     def print_formatted_text(self, formatted_text: StyleAndTextTuples) -> None:
         print_formatted_text(
             FormattedText(formatted_text),
@@ -259,7 +345,7 @@ class PythonRepl(PythonInput):
             output=self.app.output,
         )
 
-    async def _print_paginated_formatted_text(
+    def print_paginated_formatted_text(
         self, formatted_text: StyleAndTextTuples
     ) -> None:
         """
@@ -286,18 +372,30 @@ class PythonRepl(PythonInput):
             columns_in_buffer = 0
             rows_in_buffer = 0
 
-        async def show_pager() -> None:
+        def show_pager() -> None:
             nonlocal abort, max_rows
 
-            continue_result = await pager_prompt.prompt_async()
-            if continue_result == PagerResult.ABORT:
+            # Run pager prompt in another thread.
+            # Same as for the input. This prevents issues with nested event
+            # loops.
+            pager_result = None
+
+            def in_thread() -> None:
+                nonlocal pager_result
+                pager_result = pager_prompt.prompt()
+
+            th = threading.Thread(target=in_thread)
+            th.start()
+            th.join()
+
+            if pager_result == PagerResult.ABORT:
                 print("...")
                 abort = True
 
-            elif continue_result == PagerResult.NEXT_LINE:
+            elif pager_result == PagerResult.NEXT_LINE:
                 max_rows = 1
 
-            elif continue_result == PagerResult.NEXT_PAGE:
+            elif pager_result == PagerResult.NEXT_PAGE:
                 max_rows = size.rows - 1
 
         # Loop over lines. Show --MORE-- prompt when page is filled.
@@ -312,7 +410,7 @@ class PythonRepl(PythonInput):
                         # wrapping.
                         if rows_in_buffer + 1 >= max_rows:
                             flush_page()
-                            await show_pager()
+                            show_pager()
                             if abort:
                                 return
 
@@ -324,7 +422,7 @@ class PythonRepl(PythonInput):
 
             if rows_in_buffer + 1 >= max_rows:
                 flush_page()
-                await show_pager()
+                show_pager()
                 if abort:
                     return
             else:
@@ -338,9 +436,9 @@ class PythonRepl(PythonInput):
         """
         Create pager --MORE-- prompt.
         """
-        return create_pager_prompt(self._current_style)
+        return create_pager_prompt(self._current_style, self.title)
 
-    def _handle_exception(self, e: Exception) -> None:
+    def _handle_exception(self, e: BaseException) -> None:
         output = self.app.output
 
         # Instead of just calling ``traceback.format_exc``, we take the
@@ -389,6 +487,25 @@ class PythonRepl(PythonInput):
         output.write("\rKeyboardInterrupt\n\n")
         output.flush()
 
+    def _add_to_namespace(self) -> None:
+        """
+        Add ptpython built-ins to global namespace.
+        """
+        globals = self.get_globals()
+
+        # Add a 'get_ptpython', similar to 'get_ipython'
+        def get_ptpython() -> PythonInput:
+            return self
+
+        globals["get_ptpython"] = get_ptpython
+
+    def _remove_from_namespace(self) -> None:
+        """
+        Remove added symbols from the globals.
+        """
+        globals = self.get_globals()
+        del globals["get_ptpython"]
+
 
 def _lex_python_traceback(tb):
     " Return token list for traceback string. "
@@ -399,7 +516,11 @@ def _lex_python_traceback(tb):
 def _lex_python_result(tb):
     " Return token list for Python string. "
     lexer = PythonLexer()
-    return lexer.get_tokens(tb)
+    # Use `get_tokens_unprocessed`, so that we get exactly the same string,
+    # without line endings appended. `print_formatted_text` already appends a
+    # line ending, and otherwise we'll have two line endings.
+    tokens = lexer.get_tokens_unprocessed(tb)
+    return [(tokentype, value) for index, tokentype, value in tokens]
 
 
 def enable_deprecation_warnings() -> None:
@@ -472,6 +593,8 @@ def embed(
     :param configure: Callable that will be called with the `PythonRepl` as a first
                       argument, to trigger configuration.
     :param title: Title to be displayed in the terminal titlebar. (None or string.)
+    :param patch_stdout:  When true, patch `sys.stdout` so that background
+        threads that are printing will print nicely above the prompt.
     """
     # Default globals/locals
     if globals is None:
@@ -528,7 +651,9 @@ class PagerResult(Enum):
     NEXT_PAGE = "NEXT_PAGE"
 
 
-def create_pager_prompt(style: BaseStyle) -> PromptSession[PagerResult]:
+def create_pager_prompt(
+    style: BaseStyle, title: AnyFormattedText = ""
+) -> PromptSession[PagerResult]:
     """
     Create a "continue" prompt for paginated output.
     """
@@ -558,13 +683,18 @@ def create_pager_prompt(style: BaseStyle) -> PromptSession[PagerResult]:
     style
 
     session: PromptSession[PagerResult] = PromptSession(
-        HTML(
-            "<status-toolbar>"
-            "<more> -- MORE -- </more> "
-            "<key>[Enter]</key> Scroll "
-            "<key>[Space]</key> Next page "
-            "<key>[q]</key> Quit "
-            "</status-toolbar>: "
+        merge_formatted_text(
+            [
+                title,
+                HTML(
+                    "<status-toolbar>"
+                    "<more> -- MORE -- </more> "
+                    "<key>[Enter]</key> Scroll "
+                    "<key>[Space]</key> Next page "
+                    "<key>[q]</key> Quit "
+                    "</status-toolbar>: "
+                ),
+            ]
         ),
         key_bindings=bindings,
         erase_when_done=True,
